/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package mcts;

import java.util.ArrayList;
import java.util.Random;
import tictactoecodingame.Coup;
import tictactoecodingame.Joueur;
import tictactoecodingame.Plateau;

/**
 *
 * @author Lloyd
 */
public class MCTS {
    
    Plateau plateau;
    Noeud racine;
    double coefficientUCT;
    Joueur joueur;
    Joueur adversaire;
    Random r;
    
    //CONSTRUCTEUR
    public MCTS(Plateau plateau, Joueur joueur, Joueur adversaire) {
        this.plateau = plateau;
        this.joueur = joueur;
        this.adversaire = adversaire;
        
        plateau.sauvegardePosition(0);
        //coefficientUCT = 0.8;
        coefficientUCT = Math.sqrt(2);
        r = new Random(System.currentTimeMillis());
        
        
        Etat etat = new Etat();
        etat.setJoueur(joueur);
        racine = new Noeud();
        racine.setEtat(etat);
    }

    
    //Obtenir le meilleurCoup grâce au MCTS
    public Coup meilleurCoup() {
        //double endTime = System.currentTimeMillis() + 1000; // temps en milliseconde
        Noeud noeudPrometteur;
        Noeud noeudSimulation;
        Joueur gagnantSimulation;
        ArrayList<Coup> coupsPossibles;
        
        //while (System.currentTimeMillis() < endTime) {
        for (int i=0; i < 2000; i++) {    //nombre d'itérations variable
            noeudPrometteur = selection();
            
            if (plateau.partieTerminee()) {
                propagationArriere(noeudPrometteur, plateau.vainqueur());
            }
            else {
                coupsPossibles = plateau.getListeCoups(noeudPrometteur.getEtat().getJoueur());
                noeudSimulation = expansion(noeudPrometteur, coupsPossibles);
                gagnantSimulation = simulation(noeudSimulation);
                propagationArriere(noeudSimulation, gagnantSimulation);
            }
            
            /* //TESTS
            if (!racine.getListeEnfant().isEmpty()) {
                for (Noeud enfant : racine.getListeEnfant()) {
                    System.out.print(enfant.getEtat().getScoreVictoire() + " / " + enfant.getEtat().getNbVisite() + " " + enfant.getEtat().getDernierCoup().toString() + " | ");
                }
            }
            System.out.println();
            System.out.println(racine.getEtat().getScoreVictoire() + "/" + racine.getEtat().getNbVisite());
            */
            
        }
        plateau.restaurePosition(0);
        return Util.getMeilleurCoup(racine);
    }
    
    //4 PHASE DU MCTS
    
    //Selectionne le noeud "prometteur" grâce à UCT.
    public Noeud selection() {
        Noeud noeudPrometteur = racine;
        while (!noeudPrometteur.getListeEnfant().isEmpty()) {
            noeudPrometteur = UCT.trouverNoeudPrometteur(noeudPrometteur, coefficientUCT);
            plateau.joueCoup(noeudPrometteur.getEtat().getDernierCoup());
        }
        return noeudPrometteur;
    }
    
    //Une fois le meilleur noeud trouvé, on créé tous les noeuds enfant possible, et on en choisit un au hasard.
    public Noeud expansion(Noeud noeud, ArrayList<Coup> coupsPossibles) {
        Joueur joueurTemp = Util.swapJoueur(noeud.getEtat().getJoueur(), joueur, adversaire);
        int newProfondeur = noeud.getEtat().getProfondeur() + 1;
        int randomIndex;
        Etat newEtat;
        Noeud randomNoeud;
        Noeud newNoeud;
        
        for (Coup coup : coupsPossibles) {
            newEtat = new Etat();
            newEtat.setJoueur(joueurTemp);
            newEtat.setDernierCoup(coup);
            newEtat.setProfondeur(newProfondeur);

            newNoeud = new Noeud(newEtat, noeud);
            noeud.getListeEnfant().add(newNoeud);
        }
        
        
        ArrayList<Noeud> noeudsPossibles = noeud.getListeEnfant();
        randomIndex = r.nextInt(noeudsPossibles.size());
        randomNoeud = noeudsPossibles.get(randomIndex);
        plateau.joueCoup(randomNoeud.getEtat().getDernierCoup());
        return randomNoeud;
    }
    
    //A partir du nouveau noeud choisi durant l'expansion, on simule une partie de coup aléatoire jusqu'à terminer la partie.
    public Joueur simulation(Noeud noeud) {
        Joueur joueurTemp = noeud.getEtat().getJoueur();
        
        while (!plateau.partieTerminee()) {
            int randomIndex = r.nextInt(plateau.getListeCoups(joueurTemp).size());
            Coup randomCoup = plateau.getListeCoups(joueurTemp).get(randomIndex);
            plateau.joueCoup(randomCoup);
            joueurTemp = Util.swapJoueur(joueurTemp, joueur, adversaire);
        }
        
        joueurTemp = plateau.vainqueur();
        return joueurTemp;
    }
    
    //Une fois la simulation terminée, on met à jour l'arbre.
    public void propagationArriere(Noeud noeud, Joueur gagnant) {
        plateau.restaurePosition(0);
        Noeud parent = noeud;
        
        while (parent != null) {
            parent.getEtat().updateNbVisite();
            if (gagnant == null) {
                parent.getEtat().updateNbVictoire(0.5);
            }
            if (gagnant != null && parent.getEtat().getJoueur() != gagnant) {
                parent.getEtat().updateNbVictoire(1);
            }
            parent = parent.getParent();
        }
    }
}

